/*
    dp : An implementation of the davis puttnam procedure for solving
         propositional satisfiability problems.
    Copyright (C) 1993  Peter Barth

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ------------------------------------------------------------------
   DP.C
* -------------------------------------------------------------------
* Auth: pb
* Purp: A Davis Puttnam procedure implementation
* Hist:==============================================================
* Date         Name    What has been done ?!
* -------------------------------------------------------------------
* 20.8.92      pb      creation                
* 20.1.93      pb      faster,cleaned up        
* 10.4.93      pb      single linked lists for main loop
* ---------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "sys/times.h"
#include "pblist.h"
#include "silist.h"

/* To avoid warnings */
#include "protos.h"


#define max(A,B) ((A)>(B) ? (A) : (B))

#ifdef VERBOSE
#define DO_VERBOSE(V,T) if (verbosity >= V) {T}
#else
#define DO_VERBOSE(V,T)
#endif

/******************************************************/
/* The datastructures                                 */
/******************************************************/

/* The data how it is read in */
/* a matrix of integers
 *             (Ni Cij)
 * each Cij represents the number of the propositional variable
 * if Cij is negativ then it is the negated occurence
 * Ni represents the number of Literals
 * stored as list of list of ints
 */

list problem = NULL;


/* The internal datastructure */

/* a) static datastructure 
      The static datastructure is a index structure which allows to
      find all variables in a clause if the clause_id is given (the obvious one)
      and to find all clauses (clause_id's) containing a variable with a specific 
      sign if the variable and the sign is given.
 */
typedef struct vref {
  int variable;
  list clauses;
} tvref;
list variable_references = NULL;
int** array_vr = NULL;    /* The variable references as array for faster acces */
list* clauseid_vr = NULL; /* the clause_id references */

/* b) dynamic datastructure 
      The dynamic datastructure consists of two single linked lists containing
      i)  all "active variables", that is variables where a value is not yet assigned
      ii) all "active clauses", that is clause_id's which are not yet satisfied
          and the number of active variables in them (length).
      The information fits in a single integer. This implies that the number of
      clauses and the clause length is limited to 65536.
 */
#define GETtaclauselen(i) ((i) & 0x0000FFFF) 
#define GETtaclauseid(i)  ((i) >> 16) 
#define SETtaclauselen(i,v) ((i) = ((i) & 0xFFFF0000) | v) 
#define SETtaclauseid(i,v)  ((i) = ((i) & 0x0000FFFF) | ((v) << 16)) 

list active_clauses = NULL;     /* int list */
list active_variables = NULL;   /* int list */

/* Some global Flags and statistic information */
int verbosity = 1;
int timing    = 0;
int heuristic = 0;
int no_nodes = 0;
int no_ures = 0;


/******************************************************/
/* Prototypes                                         */
/******************************************************/
void write_a_v(silist);
void write_a_c(silist);


/******************************************************/
/* misc. functions                                    */
/******************************************************/

void error(char *s)
{
  fprintf(stderr,"\n%s : aborted",s);
  exit(-1);
}

int time_init_ok = 0;
struct tms ref_time;

void init_time()
{
  time_init_ok = 1;
  times(&ref_time);
}

int get_time()
{
  struct tms t;
  int time;

  if (time_init_ok == 0)
    init_time();
  times(&t);
  time = (int) ( ((long) t.tms_utime) - ((long) ref_time.tms_utime));
#ifdef linux
  return(time * 10); /* HZ is 100 */
#else
  return(time * 1000 / 60); /* normally HZ is 60 */
#endif
}

/* converts a double linked list to a single linked list assuming that the
 * data is an integer.
 */
silist list2silist(list l)
{
  silist si=init_silist();
  silist si_act = si;

  l = l_first(l);
  while(l != NULL)
    {
      si_act = si_insafter(si_act,(int) l->data);
      l = l_next(l);
    }
  return si;
}


/******************************************************/
/* INITIALIZATION                                     */
/******************************************************/

int uvr_equal(void* d1,void* d2)
{
  return ( ((tvref *)d1)->variable == ((tvref *)d2)->variable );
}

void update_variable_references(int lit,int clause)
{
  tvref svref;
  list l;

  svref.variable = abs(lit);
  svref.clauses = NULL;
  l = l_search(l_first(variable_references),(void *) &svref,uvr_equal);
  
  if (l == NULL) /* variable not yet in list */
  {
    tvref* pvref = malloc(sizeof(tvref));
    
    pvref->variable = abs(lit);
    pvref->clauses = NULL;
    pvref->clauses = l_dinsa(pvref->clauses,
                            (void *) (lit > 0 ? clause : -clause));
    variable_references = l_dinsa(l_last(variable_references),pvref);
    /* initalise active_variables */
    active_variables = l_dinsa(active_variables,(void *) abs(lit));
  }
  else
  {
    tvref* pvref = (tvref *) l_rdata(l);
    pvref->clauses = l_dinsa(pvref->clauses,    
                             (void *) (lit > 0 ? clause : -clause));
  }  
}

void init_array_vr()
{
  list vr = l_first(variable_references);
  int  max_index = 0;
  
  while(vr)
  {
    tvref *pvref = l_rdata(vr);

    max_index = max(max_index,pvref->variable);
    vr = l_next(vr);
  }
  vr = l_first(variable_references);
  array_vr = malloc(sizeof(int *) * (max_index + 2));
  if (array_vr == NULL)
    error("out of memory");
  array_vr[max_index+1] = NULL;
  while (vr)
  {
    list clauses = l_first(((tvref *)l_rdata(vr))->clauses);
    int  index   = ((tvref *)l_rdata(vr))->variable;
    int c_len = l_len(clauses);
    int j=0;

    array_vr[index] = malloc(sizeof(int) * (c_len + 1));
    if (array_vr[index] == NULL)
      error("out of memory");
    *((array_vr[index]) + c_len) = 0;
    while (clauses)
    {
      *((array_vr[index]) + j++) = (int) l_rdata(clauses);
      clauses = l_next(clauses);
    }
    vr = l_next(vr);
  }
}

int intcmp(void *d1,void *d2)
{
  return (abs((int) d1) > abs((int) d2));
}

void init_datastructures()
{
  list p = l_first(problem);
  int no_clauses = l_len(p);
  int clause_id;
  int paclause = 0;  /* to avoid warning from gcc -Wall */
  list vr;

  clauseid_vr = malloc(sizeof(list)*(no_clauses+2));

  while(p)
  {
    list clause = (list) l_rdata(p);
    clause = l_first(clause);
    clause_id = (int) l_rdata(clause);
    *(clauseid_vr+clause_id)=l_next(clause);
    SETtaclauseid(paclause,clause_id); 
    SETtaclauselen(paclause,l_len(l_next(clause))); 
    active_clauses = l_dinsa(active_clauses,(void *)paclause);
    clause = l_next(clause);
    while(clause)
    {
       int lit = (int) l_rdata(clause);

       update_variable_references(lit,clause_id);
       clause = l_next(clause);
    }
    p = l_next(p);  
  }
  vr = l_first(variable_references);
  while(vr)
  {
    tvref *pvref = (tvref *) l_rdata(vr);
    pvref->clauses = l_first(pvref->clauses);
    vr = l_next(vr);
  } 
  variable_references = l_first(variable_references);
  active_clauses = l_first(active_clauses);
  active_variables = l_sort(l_first(active_variables),intcmp);
  init_array_vr(); /* initialise the array_vr (variable references) */
}


/**************************************************************************/
/* DO THE JOB                                                             */
/**************************************************************************/

#define FAIL -1
#define OK 1
#define SATISFIABLE 2

int update(int variable,silist av,silist ac)
{
  silist av_act = av;
  silist ac_act = ac;
  int* l;
  int i;
  int ref,absref;
  int paclause;
  int absvariable = abs(variable);

DO_VERBOSE(12,printf("\nupda_start: ");write_a_v(av);printf("  ");write_a_c(ac);)
  /* delete variable from active variable list (assume av = l_first(av)) */
  while(av_act->next->key != absvariable)
     av_act = av_act->next;
  si_fastdelnext(av_act);
  av_act = av;
  /* refresh clause list */
  l = array_vr[absvariable];
  i = 0;
  /* Now l is the list of clauses where variable occurs in */
  /* assume that ac = l_first(ac); */
  do {
    paclause = ac_act->next->key;
    absref = abs(ref = l[i]);
    
    if (absref == GETtaclauseid(paclause)) /* Treffer */
    {
      /* efficient for: if ((ref > 0 && variable > 0) || (ref < 0 && variable < 0)) */
      if (!((ref ^ variable) & 0x1000))
      {
DO_VERBOSE(20,\
printf("\n    : checking %i with %i   del_clause",abs(ref),GETtaclauseid(paclause));)
	si_fastdelnext(ac_act);
      }
      else
      {
        if (GETtaclauselen(paclause) == 1)   /* empty clause derived */
	  return FAIL;
DO_VERBOSE(20, \
printf("\n    : checking %i with %i   del_literal",abs(ref),GETtaclauseid(paclause));)
	ac_act->next->key--; /* Note that len is stored on least sig bits */
        if (ac_act->next->next == ac_act->next) 
	  break;
	else 
	  ac_act = ac_act->next;
      }
      i++; /* This is updated */
    }
    else if (absref < GETtaclauseid(paclause))
      i++;
    else /* if (absref < GETtaclauseid(paclause)) */
    {
      if (ac_act->next->next == ac_act->next)
	break;
      else
	ac_act = ac_act->next;
    }
  } while (l[i]); /*  ac_act at end catched with breaks (speed) */
DO_VERBOSE(12,printf("\nupda_done : ");write_a_v(av);printf("  ");write_a_c(ac);)
  return OK;
}

int find_ucv(silist av,silist ac)
{
  list clause;
  av = av->next;
  ac = ac->next;

  for(;;) 
    if (ac->next == ac)                      /* no unit clause */
      return 0;
    else if (GETtaclauselen(ac->key)  == 1)  /* got one */
      break;
    else
      ac = ac->next;
DO_VERBOSE(11,printf("\n  : unit_clause is %i",GETtaclauseid(ac->key));)
  /* clause is the clause we search for */
  clause = *(clauseid_vr+GETtaclauseid(ac->key)); 
  do {  /* assume clause_variables and active variables to be sorted */
    int avkey = av->key;  
    int vcl = (int) clause->data; /* assume clause is valid */

    if (avkey == abs(vcl))
      return(vcl);
    else if (avkey < abs(vcl))
      av = av->next;
    else /* avkey > abs(vcl) */
      clause = clause->next; 
  } while ((av->next != av) && clause); 
#ifdef CHECK
  error("Oh what a pitty, a BUG !!!");
#endif
  exit(-1);
}

int unit_resolution(silist av,silist ac)
{
  int unit_clause_variable;
  
  do {
DO_VERBOSE(10,printf("\nures:");write_a_v(av);printf("  ");write_a_c(ac);)
    if ((unit_clause_variable = find_ucv(av,ac)) == 0)
      break;
    else
    {
      if (verbosity >= 3)
        printf("u");
      no_ures++; /* statistics */
      if (update(unit_clause_variable,av,ac) == FAIL)
        return FAIL;
    }
  } while(1); /* (ac->next != ac->next->next); (implicit because find_ucv will fail) */
DO_VERBOSE(10,printf("\nures_done");)
    if (ac->next == ac->next->next)
      return SATISFIABLE;
    else
      return OK;
}


int heuristic0(silist av,silist ac)
{
#ifdef CHECK
  if (av->next == av->next->next)
    error("kann nicht sein");
#endif
  /* very simple heuristic -> get the first variable */
  return av->next->key;
}

/* static double twopow_table[] = {
  1.0,
  0.5,
  0.25,
  0.125,
  0.0625,
  0.03125,
  0.015625,
  0.007813,
  0.003906,
  0.001953,
  0.000977,
  0.000488,
  0.000244,
  0.000122,
  6.1035e-5,
  3.0516e-5,
  1.5259e-5,
  7.629e-6,
  3.814e-6,
  1.907e-6,
  9.53e-7
};
*/

int twopow(int y)
{
  if (y > 20 || y <= 0)
    return(1);
  return (1 << (20 - y));
}


int heuristic1(silist av,silist ac) /* Jeroslow-Wang */
{
  int avlen = si_len(av);
  int* posocc = malloc(sizeof(int) * avlen);
  int* posoccptr = posocc;
  int* negocc = malloc(sizeof(int) * avlen);
  int* negoccptr = negocc;
  int i;
  int best,vbest;
  silist iav=av->next;

  silist iac;
  int* varinclauses;

  int clause_id;
  int clauselen;

#ifdef CHECK
  if (av->next->next == av->next)
    error("kann nicht sein");
  if (posocc == NULL || negocc == NULL)
    error("out of memory");
#endif
  /* calculate heuristic data_structure */
  for (i=0;i<avlen;i++) { posocc[i] = 0; negocc[i] = 0; }
  while(iav->next != iav)
  {
    varinclauses = array_vr[iav->key];
    iac = ac->next;
 
    while(*varinclauses && (iac->next != iac)) 
    {
      clause_id = GETtaclauseid((int) iac->key);
      clauselen = GETtaclauselen((int) iac->key);
 
      if(clause_id == abs(*varinclauses))
      {
        if (*varinclauses > 0) 
	  *posoccptr += twopow(clauselen); 
	else
	  *negoccptr += twopow(clauselen); 
        iac = iac->next;
        varinclauses++;        
      }  
      else if(clause_id < abs(*varinclauses))
        iac = iac->next;
      else        
        varinclauses++;        
    }
    iav = iav->next;
    posoccptr++;
    negoccptr++;    
  }
  /* Search the best */
  best = 0;
  for(i=0;i<avlen;i++)
  {
     vbest = (best < 0 ? negocc[abs(best)] : posocc[best]);
     
     if (posocc[i] > vbest) best = i;
     if (negocc[i] > vbest) best = -i;
  }
DO_VERBOSE(13,int i;for(i=0;i<avlen;i++) printf("\n%i    %i",posocc[i],negocc[i]);)
DO_VERBOSE(13,printf("\n best is %i",best);)
  free(posocc);
  free(negocc);
  vbest = abs(best);
  if (best == 0)
    return si_dith(av,1);
  else if (best > 0)
    return si_dith(av,vbest);
  else
    return -si_dith(av,vbest);
}

int heuristic2(silist av,silist ac) /* Clause with most often occurence */
{
  silist iav = av->next;
  silist iac;
  silist maxav = av;

  int* varinclauses;
  int clause_id,clauselen;
  int no_occs = 0;
  int max_no_occs = 0;
  
#ifdef CHECK
  if (iav->next == iav)
    error("should not be");
  if (posocc == NULL || negocc == NULL)
    error("out of memory");
#endif
  while(iav->next != iav)
  {
    varinclauses = array_vr[iav->key];
    iac = ac->next;
    no_occs = 0;

    while(*varinclauses && (iac->next != iac)) 
    {
      clause_id = GETtaclauseid(iac->key);
      clauselen = GETtaclauselen(iac->key);
 
      if(clause_id == abs(*varinclauses))
      {
	no_occs++;	
        iac = iac->next;
        varinclauses++;        
      }  
      else if(clause_id < abs(*varinclauses))
        iac = iac->next;
      else        
        varinclauses++;        
    }
    if (no_occs > max_no_occs) 
      { maxav = iav; max_no_occs = no_occs; }
    iav = iav->next;
  }
  return maxav->key;
}

int choose_variable(silist av,silist ac)
{
  switch(heuristic) {
  case 0  : return heuristic0(av,ac);
            break;
  case 1  : return heuristic1(av,ac);
            break;
  case 2  : return heuristic2(av,ac);
            break;
  default : error("unknown heuristic");exit(-1);
  } 
}


int solve(silist active_variables,silist active_clauses)
{
  silist cp_av = NULL;
  silist cp_ac = NULL;
  int choice_variable;
  int ret;

DO_VERBOSE(10,printf("\nsolv:");write_a_v(active_variables);printf("  ");write_a_c(active_clauses);)
  /* do unit_resolution */
  ret = unit_resolution(active_variables,active_clauses);
  if (ret == FAIL)
    return FAIL;
  else if (ret == SATISFIABLE)
    return SATISFIABLE;

  /* copy the active lists */
  cp_av = si_fastcopy(active_variables);
  cp_ac = si_fastcopy(active_clauses);
 
  choice_variable = choose_variable(cp_av,cp_ac);
DO_VERBOSE(13,printf("\ncopy_active: ");write_a_v(cp_av);write_a_c(cp_ac);\
printf("\n org_active: ");write_a_v(active_variables);write_a_c(active_clauses);)
DO_VERBOSE(10,printf("\nchoice on %i :",choice_variable);)
  if (verbosity >= 2)
    { printf("."); fflush(stdout); }
  no_nodes++; /* statistics */
  update(choice_variable,cp_av,cp_ac);
  if (cp_ac->next == cp_ac->next->next)
    return SATISFIABLE;
  if (solve(cp_av,cp_ac) == FAIL) 
  {
    if (verbosity >= 2)
      { printf("B"); fflush(stdout); }
    free(cp_av);free(cp_ac); /* free the whole block */
    no_nodes++; /* statistics */
    update(-choice_variable,active_variables,active_clauses);
    return(solve(active_variables,active_clauses));
  }
  else 
  {    
    free(cp_av);free(cp_ac); /* free the whole block */
    return SATISFIABLE;
  }
}



/**************************************************************************/
/* Input/Output                                                           */
/**************************************************************************/

void skip_chars(FILE* file,int skip)
{
  int c = fgetc(file);
  if (c == skip)
    skip_chars(file,skip);
  else
    ungetc(c,file);
}

void skip_eoln(FILE* file)
{
  skip_chars(file,'\n');
}

int next_char_is(FILE* file,int is)
{
  int c = fgetc(file);

  ungetc(c,file);
  if (c == is)
    return 1;
  else
    return 0;
}

int feoln(FILE* file)
{
  int c = fgetc(file);

  ungetc(c,file);
  if (c == '\n')
    return 1;
  else
    return 0;
}

/* I hate to write parsers, 
 * somebody like to write a lex/yacc front ? you're welcome ! */
int get_int(FILE* file)
{
  char buffer[20];
  char*bufptr = buffer;
  int c;
  int var;
  int sign = 0;
  
  do {
    c = fgetc(file);
    if (sign == 0)      /* only for the zero */
      if (c == '-') sign = -1; else sign = 1;
    if (isalpha(c))     /* skip characters */
      c = fgetc(file);
    *bufptr++ = c;
  } while ((c >= '0' && c <= '9') || c == '-');
  ungetc(*(--bufptr),file);
  *bufptr = 0x00;
  var = atoi(buffer);
  if (var == 0)
    return sign;
  if (var >= 0)
    return var+1;
  else
    return var-1;
}

int get_variable(FILE* file)
{ 
  int variable;
  int comment = 1;

  do {
    if (next_char_is(file,'*')) /* comment */
      while (fgetc(file) != '\n'); /* forget line */
    else if (next_char_is(file,'c')) /* comment */
      while (fgetc(file) != '\n'); /* forget line */
    else if (next_char_is(file,'p')) /* comment */
      while (fgetc(file) != '\n'); /* forget line */
    else if (next_char_is(file,'%')) /* comment */
      while (fgetc(file) != '\n'); /* forget line */
    else comment = 0;
  } while (comment);
  if (next_char_is(file,'E')) /* maybe ENDDATA */
  { 
    char buffer[80];
    
    fgets(buffer,80,file);
    if (strncmp("ENDDATA",buffer,7))
    {
      while (!feof(file)) fgetc(file);
      return 0;
    }
    else
      error("Line musn't begin with 'E'");
  }
  if (next_char_is(file,'0')) /* maybe END OF PROBLEM */
  { 
    skip_chars(file,'0');
    skip_chars(file,' ');
    if (next_char_is(file,'\n')) /* it is the end of the problem */
    {
      while (!feof(file)) fgetc(file);
      return 0;
    }
    /* it is the variable 0 */
    return 0+1;  /* increase the variable by 1 */
  }
  skip_chars(file,' ');
  skip_chars(file,'|');
  skip_chars(file,' ');
  skip_chars(file,'(');   /* forget parentheses */
  skip_chars(file,')');   /* forget parentheses */
  skip_chars(file,' ');
  if (!next_char_is(file,'\n') && (!next_char_is(file,'.')))
  {
    variable = get_int(file);
    return variable;
  }
  skip_chars(file,'.');
  skip_chars(file,'\n');
  return(0);
}

void read_clauses(FILE *file)
{
  int variable;
  int clause_id = 1;
  list clause=NULL;

  while (!feof(file))
  {
    variable = get_variable(file);
    if (variable) /* NOT END OF CLAUSE */
      clause = l_dinsa(clause,(void *) variable);  
                              /* pointer same size as int !!!!! */
    else
    {
      clause = l_first(clause);
      if (l_len(clause) != 0)
      {
        clause = l_sort(clause,intcmp);
        clause = l_dinsb(clause,(void *)clause_id++);
        problem = l_dinsa(problem,clause);
      }
      clause = NULL;
    }
  }
}

void write_clause(void *data)
{
  list l = (list) data;

  l = l_first(l);
  printf("  %3i(%2i): ",(int) l_rdata(l),l_len(l_next(l)));
  l = l_next(l);
  while(l)
  {
    int lit = (int) l_rdata(l);
    printf("%i ",lit);
    l = l_next(l);
  }
  printf("\n");
}

void report_problem()
{
  if (verbosity <= 3)
    return;
  printf("\n The problem to solve (%i clauses) \n",l_len(l_first(problem)));
  l_forall(l_first(problem),write_clause);
}

void write_a_v(silist av)
{ 
  av = av->next;
  while(av->next != av) 
    { 
      printf("%i ",av->key); 
      av = av->next;
    } 
}

void write_a_c(silist ac)
{
  ac = ac->next;
  while(ac->next != ac) 
  { 
    printf("%i(%i) ",GETtaclauseid(ac->key),
                     GETtaclauseid(ac->key)); 
    ac = ac->next;
  }
}

void report_datastructures()
{
  list av = l_first(active_variables);

  if (verbosity <= 3)
    return;
  printf("\n The static variable references \n");
  while(av)
  {
    int variable = (int) l_rdata(av);
    int* refs = array_vr[variable];

    printf("\n  %3i: ",variable);
    while (*refs)
      printf("%i ",*refs++);
    av = l_next(av);
  }
  { 
    silist ac = list2silist(active_clauses);
    silist av = list2silist(active_variables);
    printf("\n The active clauses   \n   ");
    write_a_c(ac);
    free_silist(ac);
    printf("\n The active variables \n   ");
    write_a_v(av);
    free_silist(av);
  }
}


/*************************************************************************/
/*                           D P H E L P                                 */
/*************************************************************************/

void dphelp()
{
  printf("This is an implementation of the well known Davis Puttnam Procedure \n");
  printf("for solving propositional logic satisfiability problems\n");
  printf("The following options are available \n");
  printf("\n");
  printf("  -H       prints this message \n\n");
  printf("  -h<n>    chose a variable selection heuristic \n");
  printf("    -h0      no heuristic -> first occurence; first try \n");
  printf("    -h1      Jeroslow Wang heuristic\n");
  printf("    -h2      most often occurrence -> select positiv \n");
  printf("  -tn      timing 0: no timing. 1: overall cputime. 2: init <-> solve \n");
  printf("  -f<name> read input from file <name> (else from stdin) \n");
  printf("  -v<n>    verbosity level 0: nothing. 1: +init+result 2:+choice+backt. \n");
  printf("                           3: +unitres \n");
  printf("                           higher for debugging \n");
  printf("\n");
  printf("Syntax :  \n");
  printf("A positive literal is an pos. int a negative literal a neg. int\n");
  printf("A clause is a sequence of literals separated by whitespace or '|' \n");
  printf("A clause ends with '.' and <END OF LINE> (Un*x EOLN) \n");
  printf("A problem description ends with 'ENDATA' or <EOF> \n");
  printf("Characters are ignored while scanning the input\n");
  printf("\n");
  printf("Remarks: The heuristics are not very fast.\n");
  printf("         The static index-datastructure is large, but while running  \n");
  printf("         memory consumption should be acceptable.  \n");
  printf("To do:   It's a hack up to now. Needs to be cleaned. \n");
  printf("         A satisfying assignment should be displayed. \n");
  printf("          \n");
  printf(" Name   : dp \n");
  printf(" Version: 0.1 // 10.4.93 \n");
  printf(" Author : Peter Barth \n");
  printf(" Email  : barth@mpi-sb.mpg.de \n");
  printf("\n");
  exit(-1);
}


/*************************************************************************/
/*                              M A I N                                  */
/*************************************************************************/

int main(int argc,char** argv)
{
  FILE *file = NULL;
  int res;
  int tmp_time = 0;
  int start_time = 0;
  int read_time = 0;
  int init_time = 0;
  int solve_time = 0;
  int end_time = 0;
  silist av,ac;

  if (argc >= 2)
  {
    int i = 1;
    char buffer[60];

    while (argc > i) {
      strcpy(buffer,argv[i]);
      if (buffer[0] != '-')
        error("option don't begin with -");
      switch(buffer[1]) {
        case 'f' : file =fopen(buffer+2,"r");
                   if (file == NULL)
                     error("File not found");
                   break;
        case 'v' : verbosity = atoi(buffer+2);
                   break;
        case 't' : timing = atoi(buffer+2);
                   break;
        case 'h' : 
	           if (isdigit(*(buffer+2)))
		     heuristic = atoi(buffer+2);
	           else
		     error("-h?  which heuristic ?");
                   break;
	case 'H' : dphelp();
	           break;
	default  : error("unknown option");
      }
      i++;
    }
  }
  if (file == NULL)
    file = stdin;
  if (timing)
    start_time = get_time();
  read_clauses(file);
  if (timing)
    read_time = get_time()-start_time;
  if (verbosity >= 1)
    { 
      printf("\
dp version 0.1, Copyright (C) 1993 Peter Barth \n\
dp comes with ABSOLUTELY NO WARRANTY; for details look in the file COPYING. \n\
This is free software, and you are welcome to redistribute it \n\
under certain conditions; look in the file COPYING for details. \n");
      printf("type    dp -H    for some help \n");
      fflush(stdout); 
    }
  report_problem();
  if (verbosity >= 1)
    { 
      printf("\nInput completed --> Initializing datastructures ... "); 
      fflush(stdout); 
    }
  if (timing)
    tmp_time = get_time();
  init_datastructures();
  if (timing)
    init_time = get_time()-tmp_time;
  report_datastructures();
  if (verbosity >= 1)
    { 
      printf("\nInitializing datastructures completed  --> Solving ... \n"); 
      fflush(stdout); 
    }
  if (timing)
    tmp_time = get_time();
  av = list2silist(active_variables);
  ac = list2silist(active_clauses);
  res = solve(av,ac);
  free_silist(av);
  free_silist(ac);
  if (timing) 
    {
      end_time = get_time();
      solve_time = end_time-tmp_time;
    }
  if (res == FAIL) 
    {
      if (verbosity >= 1)
	printf("\n ****** INSATISFIABLE ******");
    }
  else
    if (verbosity >= 1)
      printf("\n ******  SATISFIABLE  ******");
  if (timing == 1)
    printf("\nTime used : %i msec.",end_time-start_time);
  if (timing == 2)
    printf("\nTime used (msecs) : read(%i) init(%i) solve(%i) ",
                                  read_time,init_time,solve_time);
  if (verbosity >= 1)
    printf("\nExplored nodes : %i     Number of unit resolutions : %i ",
	   no_nodes,no_ures);
  if (verbosity >= 1)
    printf("\nBye ...\n");
  return(res == FAIL ? 1 : 0);
}



